/* $Id: ncron.c 6 2023-02-20 17:11:41Z nishi $ */

#include <ncron.h>

#include <ncron/parse.h>
#include <ncron/util.h>

#include <signal.h>
#include <unistd.h>
#include <time.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>

#define NCRON_MINUTE	1 << 0
#define NCRON_HOUR	1 << 1
#define NCRON_DATE	1 << 2
#define NCRON_MONTH	1 << 3
#define NCRON_DAY	1 << 4

bool ncron_match(struct ncron_entry* entry, struct tm* tm){
	if(entry->e_reboot == 1) return false;
	int i;
	char fullflag = NCRON_MINUTE | NCRON_HOUR | NCRON_DATE | NCRON_MONTH | NCRON_DAY;
	for(i = 0; entry->e_hour_start[i] != -1; i++){
		char flag = 0;
		if(((entry->e_minute_start[i] <= tm->tm_min) && (entry->e_minute_end[i] == -2 ? (entry->e_minute_start[i] == tm->tm_min) : (tm->tm_min <= entry->e_minute_end[i]))) && ((tm->tm_min % entry->e_minute_step[i]) == 0)) flag |= NCRON_MINUTE;
		if(((entry->e_hour_start[i] <= tm->tm_hour) && (entry->e_hour_end[i] == -2 ? (entry->e_hour_start[i] == tm->tm_hour) : (tm->tm_hour <= entry->e_hour_end[i]))) && ((tm->tm_hour % entry->e_hour_step[i]) == 0)) flag |= NCRON_HOUR;
		if(((entry->e_date_start[i] <= tm->tm_mday) && (entry->e_date_end[i] == -2 ? (entry->e_date_start[i] == tm->tm_mday) : (tm->tm_mday <= entry->e_date_end[i]))) && ((tm->tm_mday % entry->e_date_step[i]) == 0)) flag |= NCRON_DATE;
		if(((entry->e_month_start[i] <= tm->tm_mon) && (entry->e_month_end[i] == -2 ? (entry->e_month_start[i] == tm->tm_mon) : (tm->tm_mon <= entry->e_month_end[i]))) && ((tm->tm_mon % entry->e_month_step[i]) == 0)) flag |= NCRON_MONTH;
		if(((entry->e_day_start[i] <= tm->tm_wday) && (entry->e_day_end[i] == -2 ? (entry->e_day_start[i] == tm->tm_wday) : (tm->tm_wday <= entry->e_day_end[i]))) && ((tm->tm_wday % entry->e_day_step[i]) == 0)) flag |= NCRON_DAY;
		if(flag == fullflag) return true;
	}
	return false;
}

void ncron_start(){
	signal(SIGCHLD, SIG_IGN);
	int pid;
	struct ncrontab* tab = ncron_parse();
	int i;
	for(i = 0; tab->tabs[i] != NULL; i++){
		if(tab->tabs[i]->e_reboot){
			if(fork() == 0){
				char** env = malloc(sizeof(*env) * 3);
				env[0] = sstrdup("test=abc");
				env[1] = sstrcat("PATH=", tab->tabs[i]->e_path);
				env[2] = NULL;
				if(0 != seteuid(tab->tabs[i]->e_uid)){
					fprintf(stderr, "setuid() failed\n");
					_exit(0);
				}
				if(0 != setegid(tab->tabs[i]->e_gid)){
					fprintf(stderr, "setgid() failed\n");
					_exit(0);
				}
				execle(tab->tabs[i]->e_shell, tab->tabs[i]->e_shell, "-c", tab->tabs[i]->e_command, NULL, env);
				_exit(-1);
			}
		}
	}
	time_t now = time(0);
	struct tm* tm = localtime(&now);
	sleep(60 - (tm->tm_sec == 0 ? 60 : tm->tm_sec));	/* Wait until the next minute... */
	while(1){
	       	pid = fork();
		if(pid == 0){
			for(i = 0; tab->tabs[i] != NULL; i++){
				if(fork() == 0){
					now = time(0);
					tm = localtime(&now);
					bool matched = ncron_match(tab->tabs[i], tm);
					if(matched){
						char** env = malloc(sizeof(*env) * 3);
						env[0] = sstrdup("test=abc");
						env[1] = sstrcat("PATH=", tab->tabs[i]->e_path);
						env[2] = NULL;
						if(0 != seteuid(tab->tabs[i]->e_uid)){
							fprintf(stderr, "setuid() failed\n");
							_exit(0);
						}
						if(0 != setegid(tab->tabs[i]->e_gid)){
							fprintf(stderr, "setgid() failed\n");
							_exit(0);
						}
						execle(tab->tabs[i]->e_shell, tab->tabs[i]->e_shell, "-c", tab->tabs[i]->e_command, NULL, env);
					}
					_exit(-1);
				}
			}
			_exit(0);
		}else{
			sleep(60);
		}
	}
	ncron_ncrontab_free(tab);
}
